#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "s.h"

#define BS 4 

int cursize;
struct BBt** temp=NULL;

static void
err(char* msg)
{
	fprintf(stderr,msg);
	exit(1);
}

int
maxh(struct BBt* tree, int max, int cur)
{
	int c1, c2;	
#define MAX (cur>max?cur:max)

	if (!tree)
		return MAX;

	if (!tree->parent)
		return maxh(tree->left, max, cur);

	cur++;
	c1=maxh(tree->left, MAX, cur);
	c2=maxh(tree->right, MAX, cur);

	return (c1>c2?c1:c2);
}

static void
putspace(int n)
{
	int i;
	for (i=0;i<n;i++)
		putchar(0x20);
}

void
printBBt(struct BBt* tree)
{
	struct BBt* cur=tree->left;

	if (!cur)
		return;

	int len=(cur->wl+cur->wr)/2;
	int i, f, l, h;
	struct BBt* queue[len];
	int lev[len+1];
	int ind[len+1];
	f=l=0;	
	queue[0]=cur;
	lev[0]=0;
	memset(ind,0,sizeof(ind));
	
	while(f<=l){
		ind[l+1]+=ind[f]*2;
		if (cur->left){
			queue[++l]=cur->left;
			lev[l]=lev[f]+1;
		} else {
			ind[l+1]++;
		}
		if (cur->right){
			queue[++l]=cur->right;
			lev[l]=lev[f]+1;
		} else {
			ind[l+1]++;
		}
		cur=queue[++f];
	}

	h=lev[f-1];
	char buf[CHAR_BIT*sizeof(int)/3 + 5];
	int  spaces, carry, add;
	carry=0;
	add=1;
	for (i=0;i<f;i++)
		if ((len=sprintf(buf,"%d",queue[i]->value))>add)
			add=len;
	for (i=0; i<f; i++){
		len=sprintf(buf,"%d",queue[i]->value);
		int l1, l2;
		l2=len/2;
		l1=len-l2;

		if (i==0 || lev[i-1]<lev[i])
			spaces=(1<<(h-lev[i]))*add;
		else 
			spaces=(1<<(h-lev[i]+1))*add;
		spaces+=(1<<(h-lev[i]+1))*add*ind[i];
		if (spaces-l1-carry>0){
			putspace(spaces-l1-carry);
			carry=0;
		} else {
			putspace(1);
			carry=l1+1-spaces+carry;	
		}
		carry+=l2;
		printf("%s",buf);
		if (lev[i]<f && lev[i]!=lev[i+1]){
			carry=0;
			putchar('\n');
		}
	}
	putchar('\n');
}

static struct BBt*
newBBt(int value)
{
	struct BBt* ret;

	if (!(ret=malloc(sizeof(struct BBt))))
		err("Low memory");
	ret->value=value;
	ret->left=ret->right=ret->parent=NULL;
	ret->wl=ret->wr=1;
	return ret;

}

struct BBt*
newroot(void)
{
      	return newBBt(42); 
}


static struct BBt*
rebalance(int start, int end, struct BBt* prev)
{
	int avg;
	struct BBt* tree;

	if (start >= end)
		return NULL;
	avg=(start+end)/2;
	tree = temp[avg];
	tree->left = rebalance(start, avg, tree);
	tree->right = rebalance(avg+1, end, tree);
	tree->parent=prev;
	tree->wl = 2*(avg - start) + 1;
	tree->wr = 2*(end - (avg+1)) + 1;
	return tree;
}

static void
initBBt(void)
{
	cursize=BS;

	if (!(temp=malloc(sizeof(struct BBt*)*BS)))
		err("Low memory");
}

static int
inter_rearrange(struct BBt* tree, int i)
{
	if (!temp)
		initBBt();

	if (tree->left)
		i=inter_rearrange(tree->left,i);
	if (i>=cursize){
		cursize*=2;
		if (!(temp=realloc(temp, sizeof(struct BBt*)*cursize)))
			err("Low memory");
	}
	temp[i++]=tree;
	if (tree->right)
		i=inter_rearrange(tree->right,i);

	return i;
}

static struct BBt*
rearrange(struct BBt* tree)
{
	int j;

	if (!tree)
		return tree;

	j=inter_rearrange(tree,0);
	return rebalance(0,j,tree->parent);
}

static struct BBt*
insertBBt(struct BBt* new, struct BBt* tree, int rebalance, struct BBt* prev)
{
	int prev_rebalance = rebalance;
	
	if (!tree){
		new->parent=prev;
		return new;
	}
	if (!tree->parent){
		tree->left=insertBBt(new,tree->left,0,tree);
		return tree;
	}
	if (new->value<=tree->value){
		tree->wl+=2;
		if (tree->wl > 2*tree->wr)
			rebalance = 1;
		tree->left=insertBBt(new,tree->left,rebalance, tree);
	} else {
		tree->wr+=2;
		if (tree->wr > 2*tree->wl)
			rebalance = 1;
		tree->right=insertBBt(new,tree->right,rebalance, tree);
	}

	if (rebalance != prev_rebalance)
		tree = rearrange(tree);

	return tree;
}

struct BBt*
searchBBt(int value, struct BBt* tree)
{
	if (!tree)
		return NULL;
	if (!tree->parent)
		return searchBBt(value,tree->left);

	if (tree->value==value)
		return tree;
	else if (value<tree->value)
		return searchBBt(value,tree->left);
	else
		return searchBBt(value,tree->right);

}

void
dumpBBt(struct BBt *t, int level)
{
	int i;

	if (t && !t->parent)
		t=t->left;

	if (t)
		dumpBBt(t->right,level+1);
	for (i=0; i<level; i++)
		putchar('\t');
	if (t) {
		printf("%d (%d,%d)\n", t->value, t->wl, t->wr);
		dumpBBt(t->left,level+1);
	} else
		puts("NULL");
}

static int
reconnect(struct BBt* node, struct BBt* old, struct BBt* new)
{
	int ret=0;

	if (node->left==old){
		node->left=new;
		ret=1;
	}
	else
		node->right=new;
	if (new)
		new->parent=node;
	return ret; 

}

struct BBt*
deleteBBt(struct BBt* tree, int value)
{
	struct BBt* found, * temp, * rebalance;
	int side;

	found=searchBBt(value,tree);

	if (!found)
		return tree;	
	if (!found->left && !found->right)
		side=reconnect(found->parent,found,NULL);
	else if (!found->left && found->right)
		side=reconnect(found->parent,found,found->right);
	else if (found->left && !found->right)
		side=reconnect(found->parent,found,found->left);
	else {
		temp=found->right;
		while (temp && temp->left)
			temp=temp->left;
		found->value=temp->value;
		found=temp;

		side=reconnect(found->parent,found,found->right);
	}
	temp=found->parent;
	free(found);
	rebalance=NULL;

	while (temp->parent) {

		if (side) temp->wl-=2;
		else temp->wr-=2;

		if (temp->wl*2<temp->wr || temp->wr*2<temp->wl)
			rebalance=temp;	

		side=reconnect(temp->parent,temp,temp);		
		temp=temp->parent;
	}
	if (!rebalance)
		return tree;

	temp = rebalance->parent;
	reconnect(temp,rebalance,rearrange(rebalance));
	return tree;
}

struct BBt*
insBBt(struct BBt* tree,int value)
{
	return insertBBt(newBBt(value),tree,0, NULL);
}
